""" Package constants """

__author__ = "Michal Stolarczyk"
__email__ = "michal.stolarczyk@nih.gov"

# Project-related
NAME_KEY = "name"
CONFIG_VERSION_KEY = "pep_version"
DESC_KEY = "description"
CONFIG_FILE_KEY = "_config_file"
SAMPLE_TABLE_FILE_KEY = "_sample_table_path"
SUBSAMPLE_TABLES_FILE_KEY = "_subsample_tables_path"
SAMPLE_TABLE_INDEX_KEY = "sample_table_index"
SUBSAMPLE_TABLE_INDEX_KEY = "subsample_table_index"
CONFIG_KEY = "_config"
PROJECT_TYPENAME = "Project"
SAMPLE_MODS_KEY = "sample_modifiers"
PROJ_MODS_KEY = "project_modifiers"
NAME_TABLE_ATTR = "sample_table"
CONSTANT_KEY = "append"
REMOVE_KEY = "remove"
DUPLICATED_KEY = "duplicate"
DERIVED_KEY = "derive"
DERIVED_SOURCES_KEY = "sources"
DERIVED_ATTRS_KEY = "attributes"
IMPLIED_KEY = "imply"
IMPLIED_THEN_KEY = "then"
IMPLIED_IF_KEY = "if"
IMPLIED_COND_KEYS = [IMPLIED_IF_KEY, IMPLIED_THEN_KEY]
METADATA_KEY = "metadata"
OUTDIR_KEY = "output_dir"
CFG_IMPORTS_KEY = "import"
AMENDMENTS_KEY = "amend"
ACTIVE_AMENDMENTS_KEY = "_" + AMENDMENTS_KEY
SAMPLE_EDIT_FLAG_KEY = "_samples_touched"
SAMPLE_MODIFIERS = [CONSTANT_KEY, IMPLIED_KEY, DERIVED_KEY, DUPLICATED_KEY, REMOVE_KEY]
PROJECT_MODIFIERS = [CFG_IMPORTS_KEY, AMENDMENTS_KEY]
PROJECT_CONSTANTS = [
    "REQUIRED_VERSION",
    "CONSTANT_KEY",
    "DERIVED_SOURCES_KEY",
    "DERIVED_KEY",
    "SAMPLE_MODS_KEY",
    "IMPLIED_KEY",
    "METADATA_KEY",
    "NAME_TABLE_ATTR",
    "OUTDIR_KEY",
    "CONFIG_FILE_KEY",
    "SAMPLE_TABLE_FILE_KEY",
    "SUBSAMPLE_TABLES_FILE_KEY",
    "SAMPLE_TABLE_INDEX_KEY",
    "SUBSAMPLE_TABLE_INDEX_KEY",
    "CONFIG_KEY",
    "NAME_KEY",
    "REMOVE_KEY",
    "AMENDMENTS_KEY",
    "PROJECT_TYPENAME",
    "SAMPLE_MODS_KEY",
    "CONFIG_VERSION_KEY",
    "DUPLICATED_KEY",
    "SAMPLE_EDIT_FLAG_KEY",
    "CFG_IMPORTS_KEY",
    "ACTIVE_AMENDMENTS_KEY",
    "DERIVED_ATTRS_KEY",
    "IMPLIED_COND_KEYS",
    "IMPLIED_IF_KEY",
    "IMPLIED_THEN_KEY",
    "PROJ_MODS_KEY",
    "PROJECT_MODIFIERS",
    "DESC_KEY",
]

# Sample-related
SAMPLE_YAML_FILE_KEY = "yaml_file"
SAMPLE_YAML_EXT = (".yaml", ".yml")
SAMPLE_NAME_ATTR = "sample_name"
SUBSAMPLE_NAME_ATTR = "subsample_name"
SAMPLE_SHEET_KEY = "sample_sheet"
SUBSAMPLE_SHEET_KEY = "subsample_sheet"
CFG_SAMPLE_TABLE_KEY = "sample_table"
CFG_SUBSAMPLE_TABLE_KEY = "subsample_table"
SAMPLE_DF_KEY = "_sample_df"
SUBSAMPLE_DF_KEY = "_subsample_df"
SAMPLE_DF_LARGE = "_large_sample_df"
PRJ_REF = "_project"
ATTR_KEY_PREFIX = "_key_"
INPUTS_ATTR_NAME = "input_attrs"
REQ_INPUTS_ATTR_NAME = "required_" + INPUTS_ATTR_NAME
PROTOCOL_KEY = "protocol"
SAMPLE_CONSTANTS = [
    "PROTOCOL_KEY",
    "SUBSAMPLE_SHEET_KEY",
    "SAMPLE_SHEET_KEY",
    "PRJ_REF",
    "SAMPLE_NAME_ATTR",
    "ATTR_KEY_PREFIX",
    "CFG_SAMPLE_TABLE_KEY",
    "CFG_SUBSAMPLE_TABLE_KEY",
    "SAMPLE_DF_KEY",
    "SUBSAMPLE_DF_KEY",
    "SAMPLE_DF_LARGE",
    "SUBSAMPLE_NAME_ATTR",
    "INPUTS_ATTR_NAME",
    "REQ_INPUTS_ATTR_NAME",
    "SAMPLE_YAML_FILE_KEY",
    "SAMPLE_YAML_EXT",
    "SAMPLE_MODIFIERS",
]

# Other
REQUIRED_VERSION = ["2", "1", "0"]
PKG_NAME = "peppy"
MAX_PROJECT_SAMPLES_REPR = 20
OTHER_CONSTANTS = [
    "MAX_PROJECT_SAMPLES_REPR",
    "PKG_NAME",
    "REQUIRED_VERSION",
]

PEP_LATEST_VERSION = "2.1.0"
SAMPLE_RAW_DICT_KEY = "_sample_dict"
SUBSAMPLE_RAW_LIST_KEY = "_subsample_list"

__all__ = PROJECT_CONSTANTS + SAMPLE_CONSTANTS + OTHER_CONSTANTS



""" Configuration for modules with independent tests of models. """

import json
import os

import pandas as pd
import pytest

__author__ = "Michal Stolarczyk"
__email__ = "michal.stolarczyk@nih.gov"

# example_peps branch, see: https://github.com/pepkit/example_peps
EPB = "master"


def merge_paths(pep_branch, directory_name):
    return os.path.join(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
        "tests",
        "data",
        "example_peps-{}".format(pep_branch),
        "example_{}".format(directory_name),
    )


def get_path_to_example_file(branch, directory_name, file_name):
    return os.path.join(merge_paths(branch, directory_name), file_name)


@pytest.fixture
def example_pep_cfg_path(request):
    return get_path_to_example_file(EPB, request.param, "project_config.yaml")


@pytest.fixture
def example_pep_csv_path(request):
    return get_path_to_example_file(EPB, request.param, "sample_table.csv")


@pytest.fixture
def example_pep_nextflow_csv_path():
    return get_path_to_example_file(EPB, "nextflow_taxprofiler_pep", "samplesheet.csv")


@pytest.fixture
def example_pep_cfg_noname_path(request):
    return get_path_to_example_file(EPB, "noname", request.param)


@pytest.fixture
def example_peps_cfg_paths(request):
    """
    This is the same as the ficture above, however, it lets
    you return multiple paths (for comparing peps). Will return
    list of paths.
    """
    return [
        get_path_to_example_file(EPB, p, "project_config.yaml") for p in request.param
    ]


@pytest.fixture
def config_with_pandas_obj(request):
    return pd.read_csv(
        get_path_to_example_file(EPB, request.param, "sample_table.csv"), dtype=str
    )



"""Project configuration, particularly for logging.

Project-scope constants may reside here, but more importantly, some setup here
will provide a logging infrastructure for all of the project's modules.
Individual modules and classes may provide separate configuration on a more
local level, but this will at least provide a foundation.

"""

from ._version import __version__
from .const import *
from .exceptions import *
from .project import Project
from .sample import Sample

__classes__ = ["Project", "Sample"]
__all__ = __classes__ + ["PeppyError"]

LOGGING_LEVEL = "INFO"



from collections.abc import MutableMapping


class SimpleAttMap(MutableMapping):
    """
    Simplified the AttrMap class, which enables storing key-value pairs in
    a dictionary-like structure.
    It allows assigning and accessing the values both through attributes and items.
    In most cases used as SuperClass.
    """

    def __init__(self):
        super(SimpleAttMap, self).__init__()
        super(SimpleAttMap, self).__setattr__("_mapped_attr", {})

    def __delitem__(self, key):
        value = self[key]
        del self._mapped_attr[key]
        self.pop(value, None)

    def __setitem__(self, item, value):
        self._try_touch_samples()
        self._mapped_attr[item] = value

    def __getitem__(self, item):
        return self._mapped_attr[item]

    def __iter__(self):
        return iter(self._mapped_attr)

    def __len__(self):
        return len(self._mapped_attr)

    def __contains__(self, key):
        return key in list(self.keys())

    def __delattr__(self, key):
        del self[key]

    def __setattr__(self, item, value):
        self._mapped_attr[item] = value

    def __getattr__(self, item):
        return self._mapped_attr[item]



import os
from typing import Any, Dict, List

import pandas as pd

from .exceptions import InvalidSampleTableFileException


class TableParser:
    """
    Generic class for sample table parsers.

    Each parser must implement the following methods:
        - parse
        - ...
    """

    def __init__(self, path: str, exts: List[str]) -> None:
        self._path = path
        self._exts = exts
        self._table: pd.DataFrame = None
        self._pandas_kwargs: Dict[str, Any] = {
            "dtype": str,
            "index_col": False,
            "keep_default_na": False,
            "na_values": [""],
        }

    @property
    def path(self) -> str:
        """
        Return the path to the sample table
        """
        return self._path

    @property
    def extensions(self) -> List[str]:
        """
        Return the list of extensions supported by the parser
        """
        return self._exts

    @property
    def table(self) -> pd.DataFrame:
        """
        The parsed table

        returns pandas.DataFrame: the parsed sample table
        """
        return self._table if self._table is not None else self.parse()

    def validate_path(self) -> None:
        """
        Validate the path to the sample table

        Validation includes:
            - check whether extension is supported
            - check whether the file exists
        """
        if not any(self.path.endswith(ext) for ext in self.extensions):
            raise InvalidSampleTableFileException(
                f"Sample table file format not supported: {self.path}"
            )

    def parse(self) -> pd.DataFrame:
        """
        Parse the sample table
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        """
        Return a string representation of the parser
        """
        return f"<{self.__class__.__name__}(path={self.path})>"


class CSVTableParser(TableParser):
    """
    Parser for CSV sample tables
    """

    def __init__(self, path: str) -> None:
        super().__init__(path, ["csv"])

    def parse(self) -> pd.DataFrame:
        """
        Parse the sample table
        """
        self.validate_path()
        self._table = pd.read_csv(self.path, **self._pandas_kwargs)
        self._table = self._table.where(pd.notnull(self._table), None)
        return self.table


class TSVTableParser(TableParser):
    """
    Parser for TSV sample tables
    """

    def __init__(self, path: str) -> None:
        super().__init__(path, ["tsv"])

    def parse(self) -> pd.DataFrame:
        """
        Parse the sample table
        """
        self.validate_path()
        self._table = pd.read_csv(self.path, sep="\t", **self._pandas_kwargs)
        return self.table


class XLSXTableParser(TableParser):
    """
    Parser for MS Excel sample tables
    """

    def __init__(self, path: str) -> None:
        super().__init__(path, ["xlsx"])

    def parse(self) -> pd.DataFrame:
        """
        Parse the sample table
        """
        self.validate_path()
        self._table = pd.read_excel(self.path, **self._pandas_kwargs)
        return self.table


def select_parser(path: str) -> TableParser:
    """
    Select a parser based on the file extension

    :param str path: file path
    :return SampleTableParser: the selected parser
    :raises InvalidSampleTableFileException: if no parser is found for the extension
    """
    parsers_by_ext = parser_by_ext()
    ext = os.path.splitext(path)[1].split(".")[-1]
    if ext in parsers_by_ext:
        return parsers_by_ext[ext]
    raise InvalidSampleTableFileException(
        f"No parser found for extension: {ext}. "
        f"Supported sample table extensions: {list(parsers_by_ext.keys())}",
    )


def parser_by_ext() -> Dict[str, TableParser]:
    """
    Return a dict of parsers indexed by extension

    :return Dict[str, SampleTableParser]: dict of parsers indexed by extension
    """
    parsers_by_ext = {}
    for parser in [cls for cls in TableParser.__subclasses__()]:
        for ext in parser("").extensions:
            parsers_by_ext[ext] = parser
    return parsers_by_ext



""" Custom error types """

from abc import ABCMeta
from collections.abc import Iterable

__all__ = [
    "IllegalStateException",
    "InvalidSampleTableFileException",
    "PeppyError",
    "MissingAmendmentError",
    "InvalidConfigFileException",
    "SampleTableFileException",
    "RemoteYAMLError",
]


class PeppyError(Exception):
    """Base error type for peppy custom errors."""

    __metaclass__ = ABCMeta

    def __init__(self, msg):
        super(PeppyError, self).__init__(msg)


class IllegalStateException(PeppyError):
    """Occurrence of some illogical/prohibited state within an object."""

    pass


class SampleTableFileException(PeppyError):
    """Error type for invalid sample annotations file."""

    pass


class InvalidSampleTableFileException(SampleTableFileException):
    """Error type for invalid sample annotations file."""

    pass


class RemoteYAMLError(PeppyError):
    """Remote YAML file cannot be accessed"""

    pass


class MissingAmendmentError(PeppyError):
    """Error when project config lacks a requested subproject."""

    def __init__(self, amendment, defined=None):
        """
        Create exception with missing amendment request.

        :param str amendment: the requested (and missing) amendment
        :param Iterable[str] defined: collection of names of defined amendment
        """
        msg = "Amendment '{}' not found".format(amendment)
        if isinstance(defined, Iterable):
            ctx = "defined amendments(s): {}".format(", ".join(map(str, defined)))
            msg = "{}; {}".format(msg, ctx)
        super(MissingAmendmentError, self).__init__(msg)


class InvalidConfigFileException(PeppyError):
    """Error type for invalid project config file"""

    pass



""" Helpers without an obvious logical home. """

import logging
import os
from typing import Dict
from urllib.request import urlopen

import yaml
from ubiquerg import expandpath, is_url

from .const import CONFIG_KEY, SAMPLE_TABLE_INDEX_KEY, SUBSAMPLE_TABLE_INDEX_KEY
from .exceptions import *

_LOGGER = logging.getLogger(__name__)


def copy(obj):
    def copy(self):
        """
        Copy self to a new object.
        """
        from copy import deepcopy

        return deepcopy(self)

    obj.copy = copy
    return obj


def make_abs_via_cfg(maybe_relpath, cfg_path, check_exists=False):
    """Ensure that a possibly relative path is absolute."""
    if not isinstance(maybe_relpath, str):
        raise TypeError(
            "Attempting to ensure non-text value is absolute path: {} ({})".format(
                maybe_relpath, type(maybe_relpath)
            )
        )
    if os.path.isabs(maybe_relpath) or is_url(maybe_relpath):
        _LOGGER.debug("Already absolute")
        return maybe_relpath
    # Maybe we have env vars that make the path absolute?
    expanded = expandpath(maybe_relpath)
    if os.path.isabs(expanded):
        _LOGGER.debug("Expanded: {}".format(expanded))
        return expanded
    # Set path to an absolute path, relative to project config.
    config_dirpath = os.path.dirname(cfg_path)
    _LOGGER.debug("config_dirpath: {}".format(config_dirpath))
    abs_path = os.path.join(config_dirpath, maybe_relpath)
    _LOGGER.debug("Expanded and/or made absolute: {}".format(abs_path))
    if check_exists and not os.path.exists(abs_path):
        raise OSError(f"Path made absolute does not exist: {abs_path}")
    return abs_path


def grab_project_data(prj):
    """
    From the given Project, grab Sample-independent data.

    There are some aspects of a Project of which it's beneficial for a Sample
    to be aware, particularly for post-hoc analysis. Since Sample objects
    within a Project are mutually independent, though, each doesn't need to
    know about any of the others. A Project manages its, Sample instances,
    so for each Sample knowledge of Project data is limited. This method
    facilitates adoption of that conceptual model.

    :param Project prj: Project from which to grab data
    :return Mapping: Sample-independent data sections from given Project
    """
    if not prj:
        return {}

    try:
        return prj[CONFIG_KEY].to_dict()
    except KeyError:
        raise KeyError("Project lacks section '{}'".format(CONFIG_KEY))


def make_list(arg, obj_class):
    """
    Convert an object of predefined class to a list of objects of that class or
    ensure a list is a list of objects of that class

    :param list[obj] | obj arg: string or a list of strings to listify
    :param str obj_class: name of the class of intrest
    :return list: list of objects of the predefined class
    :raise TypeError: if a faulty argument was provided
    """

    def _raise_faulty_arg():
        raise TypeError(
            "Provided argument has to be a List[{o}] or a {o}, "
            "got '{a}'".format(o=obj_class.__name__, a=arg.__class__.__name__)
        )

    if isinstance(arg, obj_class):
        return [arg]
    elif isinstance(arg, list):
        if not all(isinstance(i, obj_class) for i in arg):
            _raise_faulty_arg()
        else:
            return arg
    else:
        _raise_faulty_arg()


def load_yaml(filepath):
    """
    Load a local or remote YAML file into a Python dict

    :param str filepath: path to the file to read
    :raises RemoteYAMLError: if the remote YAML file reading fails
    :return dict: read data
    """
    if is_url(filepath):
        _LOGGER.debug(f"Got URL: {filepath}")
        try:
            response = urlopen(filepath)
        except Exception as e:
            raise RemoteYAMLError(
                f"Could not load remote file: {filepath}. "
                f"Original exception: {getattr(e, 'message', repr(e))}"
            )
        else:
            data = response.read().decode("utf-8")
            return yaml.safe_load(data)
    else:
        with open(os.path.abspath(filepath), "r") as f:
            data = yaml.safe_load(f)
        return data


def is_cfg_or_anno(file_path, formats=None):
    """
    Determine if the input file seems to be a project config file (based on the file extension).
    :param str file_path: file path to examine
    :param dict formats: formats dict to use. Must include 'config' and 'annotation' keys.
    :raise ValueError: if the file seems to be neither a config nor an annotation
    :return bool: True if the file is a config, False if the file is an annotation
    """
    formats_dict = formats or {
        "config": (".yaml", ".yml"),
        "annotation": (".csv", ".tsv"),
    }
    if file_path is None:
        return None
    if file_path.lower().endswith(formats_dict["config"]):
        _LOGGER.debug(f"Creating a Project from a YAML file: {file_path}")
        return True
    elif file_path.lower().endswith(formats_dict["annotation"]):
        _LOGGER.debug(f"Creating a Project from a CSV file: {file_path}")
        return False
    raise ValueError(
        f"File path '{file_path}' does not point to an annotation or config. "
        f"Accepted extensions: {formats_dict}"
    )


def extract_custom_index_for_sample_table(pep_dictionary: Dict):
    """Extracts a custom index for the sample table if it exists"""
    return (
        pep_dictionary[SAMPLE_TABLE_INDEX_KEY]
        if SAMPLE_TABLE_INDEX_KEY in pep_dictionary
        else None
    )


def extract_custom_index_for_subsample_table(pep_dictionary: Dict):
    """Extracts a custom index for the subsample table if it exists"""
    return (
        pep_dictionary[SUBSAMPLE_TABLE_INDEX_KEY]
        if SUBSAMPLE_TABLE_INDEX_KEY in pep_dictionary
        else None
    )
